package br.ufsc.uid2uml;

import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Vector;

import br.ufsc.uid2uml.UIDElements.*;
import br.ufsc.uid2uml.UMLDiagram.*;

/**
 * @author Patricia Vilain
 *
 */
public class ClassDiagramGenerator {
	
	private List<UID> setUIDs;
	private String classDiagramName;
	private ClassDiagram classDiagram;
	
	/**
	 * This method creates a generator of class diagram.
	 * All information will be included in the same class diagram.
	 *
	 */	
	public ClassDiagramGenerator (String classDiagramName){
		this.setUIDs = new ArrayList<UID>();
		this.classDiagramName = classDiagramName;
	}
	
	public void insertUID(UID uid){
		
		this.setUIDs.add(uid);		
	}
	
	public void generateClassDiagram (){
		
		this.classDiagram = new ClassDiagram (this.classDiagramName);
		
		// Identify the classes from all UIDs.
		this.identifyClasses (classDiagram);
		
		// IDENTIFICA OS ATRIBUTOS ANALISANDO TODOS OS UIDs.
		this.identifyAttributes (classDiagram);
		
		// IDENTIFICA AS ASSOCIACOES ANALISANDO TODOS OS UIDs.
		this.identifyAssociations (classDiagram);	
		
		try{
			ClassDiagramToXMIConverter.getConverter().convertClassDiagramToXMI(classDiagram);
		}
		catch (Exception e){
		}

		
	}

	// DIRETRIZ 1: para cada estrutura dos UIDs  definida uma classe.
	// Tambm so definidas classes para as estruturas internas.
		
	private void identifyClasses (ClassDiagram diagram){
		for (int k = 0; k < this.setUIDs.size(); k++){
			UID uid = this.setUIDs.get(k);
			Vector<AbstractState> states = uid.getSetStates();
			for (int i = 0; i < states.size(); i++){
				AbstractState state = states.elementAt(i);
				Vector<Structure> structures = state.getAllStructures();
				for (int j = 0, size = structures.size(); j < size; j++){
					Structure structure = structures.elementAt(j);
					this.mapStructureToClass(diagram, structure);
				}
			}
		}
	}

	private void mapStructureToClass(ClassDiagram diagram, Structure structure){
	//Metodo necessario para mapear cada estrutura e suas estruturas internas para classes.
	//Mesmo que a estrutura ja tenha sido inserida anteriormente, verifica se ela tem 
	//estruturas aninhadas e, em caso afirmativo, mapeia essas estruturas para classes tb.
		try{
			diagram.insertClass(structure.getName());
		}
		catch (ClassDiagramException e){
		}
		finally{
			List<Structure> internalStructures = structure.getStructures();
			for (int j = 0, size = internalStructures.size(); j < size; j++){
				Structure internalStruct = internalStructures.get(j);
				this.mapStructureToClass(diagram, internalStruct);
			}
		}
	}
		
	
	private void identifyAttributes(ClassDiagram classDiagram) {

			List<DataItem> dataItens = new ArrayList<DataItem>();
			List<DataItem> dataItensWithNoStructure = new ArrayList<DataItem>();
			List<InputEnumeratedData> enumeratedData = new ArrayList<InputEnumeratedData>();

			List<String> assignedAttributesNames = new ArrayList<String>();
			
			for (UID uid: this.setUIDs){
				dataItens.addAll(uid.getDataItemsOfAllStructures());
			}
			
			for(DataItem dt: dataItens){
				if(!assignedAttributesNames.contains(dt.getName())){
					boolean completed = false;
					List<String> structsNames = this.structs(dt.getName());
					if(structsNames.size() == 1){
						//item de dado pertence a somente uma estrutura, ento ele  mapeado para um 
						//atributo da classe correspondente a esta estrutura.
						//DIRETRIZ 2-A
						this.classDiagram.getClass(structsNames.get(0)).setAttribute(new Attribute(dt.getName()));
						assignedAttributesNames.add(dt.getName());
						completed = true;
					} 
					else 
						if(structsNames.size() > 1){   //item de dado pertence a vrias estruturas

							//verifica se somente em uma estrutura o item de dado no aparece como conjunto.
							//para isso sao localizadas todas as estruturas aonde o dataItem aparece com cardinalidade maxima 1.
							List<String> structsWithMaxCardinalityOne = new ArrayList<String>();
							for(String structName : structsNames){
								if(cardinality(structName, dt.getName()) == 1){
									structsWithMaxCardinalityOne.add(structName);
								}
							}
							//se o item de dado aparece em uma unica estrutura com cardinalidade max igual a 1, 
							//entao torna-se um atributo da classe correspondente a esta estrutura.
							//DIRETRIZ 2-B
							if(structsWithMaxCardinalityOne.size() == 1){
								this.classDiagram.getClass(structsWithMaxCardinalityOne.get(0)).setAttribute(new Attribute(dt.getName()));
								assignedAttributesNames.add(dt.getName());
								completed = true;
							}
							else 
								if (structsWithMaxCardinalityOne.size() > 1){
									//se o item de dado aparece em mais de uma estrutura com cardinalidade max igual a 1
									//entao torna-se um atributo da(s) classe(s) correspondente(s) a(s) estrutura(s) 
									//que possuem somente o item de dado como nico item de dado.
									//DIRETRIZ 2-C
									
									for (int i = 0; i < structsWithMaxCardinalityOne.size(); i++){
										String currentStruct = structsWithMaxCardinalityOne.get(i);
										if (structHasOnlyOneDataItem(currentStruct,dt.getName())){
											this.classDiagram.getClass(currentStruct).setAttribute(new Attribute(dt.getName()));
											assignedAttributesNames.add(dt.getName());
											completed = true;
										}
									}	
									
									if (!completed) {
									
										//se o item de dado aparece em mais de uma estrutura com cardinalidade max igual a 1, 
										//entao torna-se um atributo da(s) classe(s) correspondente(s) a(s) estrutura(s) em que ele aparece
										//associado a ela(s) em mais UIDs.
										//DIRETRIZ 2-D
										
										String maxStruct = structsWithMaxCardinalityOne.get(0);
										int maxTimes = times(maxStruct, dt.getName());
										for (int i = 1; i < structsWithMaxCardinalityOne.size(); i++){
											String anotherStruct = structsWithMaxCardinalityOne.get(i);
											int anotherTimes = times(anotherStruct, dt.getName());
											if (anotherTimes > maxTimes){
												maxTimes = anotherTimes;
												maxStruct = anotherStruct;
											}
										}
										
										for (int i = 0; i < structsWithMaxCardinalityOne.size(); i++){
											String currentStruct = structsWithMaxCardinalityOne.get(i);
											if (times(currentStruct, dt.getName()) == maxTimes){
												this.classDiagram.getClass(currentStruct).setAttribute(new Attribute(dt.getName()));
												assignedAttributesNames.add(dt.getName());
												completed = true;
											}
										}	
								}
								else {
						
									if(!completed){

										//DIRETRIZ 2-E
										
										for (String struct: structsNames){
											this.classDiagram.getClass(struct).setAttribute(new Attribute(dt.getName()));
											assignedAttributesNames.add(dt.getName());
										}
									}		
								}
					}						
				}
			}
			}
			
			br.ufsc.uid2uml.UMLDiagram.Class generalClass = classDiagram.getClass("General");
			if (generalClass == null)
				classDiagram.insertClass("General");

			for (UID uid: this.setUIDs){
				dataItensWithNoStructure.addAll(uid.getDataItemsWithNoStructure());
				enumeratedData.addAll(uid.getAllInputEnumeratedData());
			}
				
				// Itens de dados que no pertenciam a nenhuma estrutura e que, se ainda no 
				// foram mapeados para atributos, sero includos como atributos da classe General
				for (DataItem dataItemWithNoStructure : dataItensWithNoStructure){
					boolean contains = false;					
					for (String attributeName : assignedAttributesNames){
						if (dataItemWithNoStructure.getName().equals(attributeName))
							contains = true;
					}
					if (!contains){
							assignedAttributesNames.add(dataItemWithNoStructure.getName());
							classDiagram.insertAttributeInClass(dataItemWithNoStructure.getName(), "General");
					}
				}

				// Dados enumerados que estavam soltos nos estados e subestados e que, se ainda no 
				// foram mapeados para atributos, sero includos como atributos da classe General
				for (InputEnumeratedData enumData : enumeratedData){
					boolean contains = false;					
					for (String attributeName : assignedAttributesNames){
						if (enumData.getName().equals(attributeName))
							contains = true;
					}
					if (!contains){
							assignedAttributesNames.add(enumData.getName());
							classDiagram.insertAttributeInClass(enumData.getName(), "General");
					}
				}
			
			
			List<Attribute> generalAttributes = classDiagram.getAttributesOfClass("General");
			if (generalAttributes.size() == 0)
				classDiagram.removeClass("General");
			
	}

	private boolean structHasOnlyOneDataItem (String currentStruct, String nameItem) {
		//Funo que retorna true se a estrutura com nome currentStruct possui somente nico elemento.

		boolean hasOnlyOneDataItem = true;

		// para cada UID, para cada State e SubState, verifica se ele contem a estrutura e se ela tem somente 
		// um item de dado.
		for (int i = 0; i < this.setUIDs.size(); i++) {
			UID uid = this.setUIDs.get(i);
			for (AbstractState s : uid.getSetStates()) {
				for (Structure str : s.getAllStructuresAndInternalStructures()) {
					if (str.getName().equals(currentStruct)) {
						if (str.getDataItems().size()>1) {
							hasOnlyOneDataItem = false;
						}
					}
				}
			}
		}
		return hasOnlyOneDataItem;
	}
	

	private void identifyAssociations (ClassDiagram diagram){
		
		//ANALISE DOS ITENS DE DADOS QUE ESTAO DENTRO DAS ESTRUTURAS (DIRETRIZ 3)
		for (UID uid : this.setUIDs){
			Vector<AbstractState> states = uid.getSetStates();
			for (int i = 0, size=states.size(); i < size; i++){
				AbstractState state = states.elementAt(i);
				Vector<Structure> structures = state.getAllStructures();
				for (int j = 0, sizeS = structures.size(); j < sizeS; j++){
					Structure structure = structures.elementAt(j);
					this.mapDataItemsOfStructureToAssociations(diagram, structure);
				}
			}
		}

		//ANALISE DOS ESTRUTURAS INTERNAS  OUTRAS ESTRUTURAS (DIRETRIZ 4)
		for (UID uid : this.setUIDs){
			Vector<AbstractState> states = uid.getSetStates();
			for (int i = 0, size=states.size(); i < size; i++){
				AbstractState state = states.elementAt(i);
				Vector<Structure> structures = state.getAllStructures(); // pegou as estruturas e todas as estruturas internas tb.
				for (int j = 0, sizeS = structures.size(); j < sizeS; j++){
					Structure structure = structures.elementAt(j);
					this.mapInternalStructureToAssociations(diagram, structure);
				}
			}
		}
		
		//ANALISE DAS TRANSIES ENTRE OS ESTADOS (DIRETRIZ 5)
		for (UID uid : this.setUIDs){
			Vector<Transition> transitions = uid.getSetTransitions();
			for (Transition transition : transitions) {

				//No est tratando transio cujo destino  outro UID.
				if (transition.getTargetType()==Transition.UID){
					// TODO
				}

				//Destino da transio  um estado de interao do mesmo UID.
				else {
					//Encontrando a classe correspondente ao estado origem
					List<br.ufsc.uid2uml.UMLDiagram.Class> sourceClasses = this.getClassesCorrespondingToSourceState(diagram, transition.getSource());
					
					//Encontrando a classe correspondente ao estado origem
					List<br.ufsc.uid2uml.UMLDiagram.Class> targetClasses = this.getClassesMoreMeaningfulOfStateOrSubstate(diagram, transition.getTargetState());
					
					//O estado origem  diferente do estado destino, define-se uma associao
					for (br.ufsc.uid2uml.UMLDiagram.Class aSourceClass : sourceClasses){
						for (br.ufsc.uid2uml.UMLDiagram.Class aTargetClass : targetClasses) {
							if (aSourceClass != aTargetClass){
								Association a;
								if (!aSourceClass.containsAssociationWithClassAndSameAssociationName(aTargetClass.getName(), aSourceClass.getName()+"-"+aTargetClass.getName()) &&
									!aSourceClass.containsAssociationWithClassAndSameAssociationName(aTargetClass.getName(), aTargetClass.getName()+"-"+aSourceClass.getName())){
									
									a = diagram.insertAssociationBetweenClasses(aSourceClass.getName()+"-"+aTargetClass.getName(), 
											aSourceClass.getName(), aTargetClass.getName()); //aClass = source; anotherClass = target
									
									this.analisaCardinalidadeSourceTarget (aSourceClass, aTargetClass, a, transition);
									
								}
								else {	
									if (aSourceClass.containsAssociationWithClassAndSameAssociationName(aTargetClass.getName(), aSourceClass.getName()+"-"+aTargetClass.getName())){
										a = aSourceClass.getAssociationWithClassAndSameAssociationName(aTargetClass.getName(), aSourceClass.getName()+"-"+aTargetClass.getName());
										this.analisaCardinalidadeSourceTarget (aSourceClass, aTargetClass, a, transition);
									}	
									else {
										a = aSourceClass.getAssociationWithClassAndSameAssociationName(aTargetClass.getName(), aTargetClass.getName()+"-"+aSourceClass.getName()); 
										this.analisaCardinalidadeTargetSource (aSourceClass, aTargetClass, a, transition);
									}
								}	
									
							}
						}
					} 
					
				}
			}
		}
	}
	
	private void analisaCardinalidadeSourceTarget (br.ufsc.uid2uml.UMLDiagram.Class aSourceClass, br.ufsc.uid2uml.UMLDiagram.Class aTargetClass, 
			                                       Association a, Transition transition){
		if (a != null){
		
			// Se N itens de dados podem ser selecionados no estado de interao fonte, 
			// ento a cardinalidade mxima da classe correspondente ao estado de interao fonte  N.
			if (transition instanceof TransitionFromSet){
				if (((TransitionFromSet)transition).getNumberOfSelectedElements() > 1){
					a.setSourceMaxCardinality(-1);
				}	
			}
					
			// Se a estrutura ou item de dado correspondente  informao mais significativa do estado 
			// de interao destino  opcional (i.e. a estrutura ou item de dado pode estar vazio), 
			// ento a cardinalidade mnima da classe correspondente ao estado de interao destino  0.
			// Se a estrutura ou item de dado, retornado no estado de interao destino,  um conjunto, 
			// ento a cardinalidade mxima da classe correspondente ao estado de interao destino  N. 		
			
			boolean encontrou = false;
			for (Structure structure : transition.getTargetState().getAllStructures()){
				if (structure.getName().equals(aTargetClass.getName())){
					if (structure.isOptional())  //estrutura correspondente a classe destino  opcional 
						a.setTargetMinCardinality(0);
					if (structure.isSet())       //estrutura correspondente a classe destino  um conjunto
						a.setTargetMaxCardinality(-1);
					encontrou = true;
				}
			}
			if (!encontrou){
				for (DataItem dataItem : transition.getTargetState().getAllDataItemsWithNoStructures()){
					if (aTargetClass.containsAttribute(dataItem.getName())){
						if (dataItem.isOptional())	//item de dado correspondente a classe destino  opcional 
							a.setTargetMinCardinality(0);
						if (dataItem.isSet())		//item de dado correspondente a classe destino  um conjunto
							a.setTargetMaxCardinality(-1);
					}
				}
			}	
		}	
	}
	
	private void analisaCardinalidadeTargetSource (br.ufsc.uid2uml.UMLDiagram.Class transitionSourceClass, 
			br.ufsc.uid2uml.UMLDiagram.Class transitionTargetClass, Association a, Transition transition){
		if (a != null){
		
		// Este metodo eh semelhante ao metodo analisaCardinalidadeSourceTarget, mas eh necessario para os casos em 
		// que a associacao ja existe, mas as classes source e target esto invertidas em relacao a transicao
		// que esta sendo analisada.
			
		// Se N itens de dados podem ser selecionados no estado de interao source, 
		// ento a cardinalidade mxima da classe correspondente ao estado de interao source  N.
			if (transition instanceof TransitionFromSet){
				if (((TransitionFromSet)transition).getNumberOfSelectedElements() > 1){
					a.setTargetMaxCardinality(-1);
				}	
			}
					
			// Se a estrutura ou item de dado correspondente  informao mais significativa do estado 
			// de interao destino  opcional (i.e. a estrutura ou item de dado pode estar vazio), 
			// ento a cardinalidade mnima da classe correspondente ao estado de interao destino  0.
			// Se a estrutura ou item de dado, retornado no estado de interao destino,  um conjunto, 
			// ento a cardinalidade mxima da classe correspondente ao estado de interao destino  N. 		
			
			boolean encontrou = false;
			for (Structure structure : transition.getTargetState().getAllStructures()){
				if (structure.getName().equals(transitionTargetClass.getName())){
					if (structure.isOptional())  //estrutura correspondente a classe destino  opcional 
						a.setSourceMinCardinality(0);
					if (structure.isSet())       //estrutura correspondente a classe destino  um conjunto
						a.setSourceMaxCardinality(-1);
					encontrou = true;
				}
			}
			if (!encontrou){
				for (DataItem dataItem : transition.getTargetState().getAllDataItemsWithNoStructures()){
					if (transitionTargetClass.containsAttribute(dataItem.getName())){
						if (dataItem.isOptional())	//item de dado correspondente a classe destino  opcional 
							a.setSourceMinCardinality(0);
						if (dataItem.isSet())		//item de dado correspondente a classe destino  um conjunto
							a.setSourceMaxCardinality(-1);
					}
				}
			}	
		}	
	}
	
	private List<br.ufsc.uid2uml.UMLDiagram.Class> getClassesCorrespondingToSourceState(ClassDiagram diagram, Object source){
		//State, DataItem, Structure, Substate
		if (source instanceof AbstractState){ //State
			return this.getClassesMoreMeaningfulOfStateOrSubstate(diagram, source);
		}
		else 
			if (source instanceof DataItem){ //DataItem
				List<br.ufsc.uid2uml.UMLDiagram.Class> resultingClasses = new ArrayList<br.ufsc.uid2uml.UMLDiagram.Class>();
				for (br.ufsc.uid2uml.UMLDiagram.Class aClass : diagram.getClasses()){
					if (aClass.containsAttribute(((DataItem)source).getName())){
						resultingClasses.add(aClass);
					}
				}
				return resultingClasses;
			}
			else
				if (source instanceof Structure){ //Structure
					List<br.ufsc.uid2uml.UMLDiagram.Class> resultingClasses = new ArrayList<br.ufsc.uid2uml.UMLDiagram.Class>();
					resultingClasses.add(diagram.getClass(((Structure)source).getName()));
					return resultingClasses;
				}
				else 
					if (source instanceof Substate){
						return this.getClassesMoreMeaningfulOfStateOrSubstate(diagram, source);						
					}
					else {
						return null;
					}
	}

	
	
	private List<br.ufsc.uid2uml.UMLDiagram.Class> getClassesMoreMeaningfulOfStateOrSubstate (ClassDiagram diagram, Object state){
		if (state instanceof AbstractState || state instanceof Substate){
			List<br.ufsc.uid2uml.UMLDiagram.Class> resultingClasses = new ArrayList<br.ufsc.uid2uml.UMLDiagram.Class>();
			Vector<Structure> structures = new Vector<Structure>();
			if (state instanceof AbstractState) 
				structures = ((AbstractState)state).getAllStructures();
			else if (state instanceof Substate)
				structures = ((Substate)state).getAllStructures();
			
			//Se o estado de interao no apresenta nenhuma estrutura, a classe que apresentar mais atributos correspondentes 
			//aos itens de dados que aparecem na estado de interao  a informao mais significativa deste estado.
			if (structures.size() == 0){ 
				this.getClassesMoreMeaningfulFromDataItemsWithoutStructures(diagram, state, resultingClasses);
			}
			else {
				if (structures.size() == 1){ //Se o estado de interao apresenta somente 1 estrutura, a classe correspondente a esta estrutura  a informao mais significativa deste estado de interao.
					resultingClasses.add(diagram.getClass(structures.elementAt(0).getName()));
				}
				else{ //Se o estado de interao apresenta mais de uma estrutura
					
					List<Structure> structuresAreNotASet = new ArrayList<Structure>();
					for (Structure structure : structures){
						if (!structure.isSet())
							structuresAreNotASet.add(structure);
					}
	
					//2.1 Se somente uma estrutura no  um conjunto, a classe correspondente a esta 
					//estrutura  a informao mais significativa deste estado de interao;
					if (structuresAreNotASet.size()==1){ 
						resultingClasses.add(diagram.getClass(structuresAreNotASet.get(0).getName()));
					}
					else {
						// 2.2 Se duas ou mais estruturas no so conjuntos, a classe correspondente 
						//  estrutura que tem menos itens de dados que no correspondem a seus atributos 
						//  a informao mais significativa deste estado de interao. Se duas ou mais 
						// estruturas satisfazem a esta condio, a estrutura que possui mais itens de dados 
						// correspondentes ao atributos da sua classe correspondente  a informao mais 
						// significativa deste estado de interao. Se duas ou mais estruturas satisfazem 
						// as duas condies, estas estruturas sero as informaes mais significativas deste estado de interao; 
						if (structuresAreNotASet.size()>1){
							this.getClassesMoreMeaningfulFromStructures(diagram, structuresAreNotASet, resultingClasses); //estou passando somente as estruturas que no so conjunto
						}
						else{ //structuresAreNotASet.size() eh igual a 0
							// 2.3 Se todas as estruturas so conjuntos, a classe correspondente  estrutura que 
							// tem menos itens de dados que no correspondem a seus atributos  a informao mais 
							// significativa deste estado de interao. Se duas ou mais estruturas satisfazem a esta condio, 
							// a estrutura que possui mais itens de dados correspondentes ao atributos da sua classe 
							// correspondente  a informao mais significativa deste estado de interao. 
							// Se duas ou mais estruturas satisfazem as duas condies, estas estruturas sero as informaes 
							// mais significativas deste estado de interao;  
							this.getClassesMoreMeaningfulFromStructures(diagram, structures, resultingClasses); //estou passando todas as estruturas, porque todas elas so conjuntos
						}
					}				
				}
			}      
			return resultingClasses;
		}
		else return null;
	}

	private void getClassesMoreMeaningfulFromDataItemsWithoutStructures (ClassDiagram diagram, Object state, List<br.ufsc.uid2uml.UMLDiagram.Class> resultingClasses){
		if (state instanceof AbstractState || state instanceof Substate){
			List<br.ufsc.uid2uml.UMLDiagram.Class> existingClasses = diagram.getClasses();
			int numberOfClasses = existingClasses.size();
			int[] classesCounter = new int[numberOfClasses];
			Arrays.fill(classesCounter, 0);
			for (int i = 0; i<numberOfClasses; i++){
				// Analisa os itens de dados
				List<DataItem> dataItems = new ArrayList<DataItem>();
				if (state instanceof AbstractState)
					dataItems = ((AbstractState)state).getAllDataItemsWithNoStructures();
				else if (state instanceof Substate)
					dataItems = ((Substate)state).getAllDataItemsWithNoStructures();
				for (int j=0, size=dataItems.size(); j<size; j++){
					if (existingClasses.get(i).containsAttribute(dataItems.get(j).getName()))
						classesCounter[i]++;
				}
				
				//Analisa os dados enumerados	
				List<InputEnumeratedData> enumeratedData = new ArrayList<InputEnumeratedData>();
				if (state instanceof AbstractState)
					enumeratedData = ((AbstractState)state).getAllInputEnumeratedData();
				else if (state instanceof Substate)
					enumeratedData = ((Substate)state).getAllInputEnumeratedData();
				for (int j=0, size=enumeratedData.size(); j<size; j++){
					if (existingClasses.get(i).containsAttribute(enumeratedData.get(j).getName()))
						classesCounter[i]++;
				}	
			}
			int max = 0;
			for (int i = 0; i<numberOfClasses; i++){
				if (classesCounter[i]>max){
					max = classesCounter[i];
					resultingClasses.clear();
					resultingClasses.add(existingClasses.get(i));
				}
				else{
					if (classesCounter[i]==max){
						resultingClasses.add(existingClasses.get(i));
					}
				}
			}
		}	
	}

	private void getClassesMoreMeaningfulFromStructures(ClassDiagram diagram, List<Structure> structures, List<br.ufsc.uid2uml.UMLDiagram.Class> resultingClasses){
		// 2.2 Se duas ou mais estruturas no so conjuntos, a classe correspondente 
		//  estrutura que tem menos itens de dados que no correspondem a seus atributos 
		//  a informao mais significativa deste estado de interao. Se duas ou mais 
		// estruturas satisfazem a esta condio, a estrutura que possui mais itens de dados 
		// correspondentes ao atributos da sua classe correspondente  a informao mais 
		// significativa deste estado de interao. Se duas ou mais estruturas satisfazem 
		// as duas condies, estas estruturas sero as informaes mais significativas deste estado de interao; 

		//Procura os itens de dados que nao pertencem a estrutura
		int numberOfStructures = structures.size();
		int[] dataItemsThatNotBelongCounter = new int[numberOfStructures];
		Arrays.fill(dataItemsThatNotBelongCounter, 0);
		for (int i=0, size=structures.size(); i<size; i++){
			List<DataItem> dataItems = structures.get(i).getDataItems();
			for (DataItem dataItem : dataItems){
				if (!diagram.getClass(structures.get(i).getName()).containsAttribute(dataItem.getName())){
					dataItemsThatNotBelongCounter[i]++;
				}
			}
		}
		
		//Procura a estrutura que tem menos itens de dados que no correspondem a seus atributos
		List<Structure> structuresWithLessDataItemsNotCorresponding = new ArrayList<Structure>();
		int min = dataItemsThatNotBelongCounter[0];
		structuresWithLessDataItemsNotCorresponding.add(structures.get(0));
		for (int i = 1; i<numberOfStructures; i++){
			if (dataItemsThatNotBelongCounter[i]<min){
				min = dataItemsThatNotBelongCounter[i];
				structuresWithLessDataItemsNotCorresponding.clear();
				structuresWithLessDataItemsNotCorresponding.add(structures.get(i));
			}
			else{
				if (dataItemsThatNotBelongCounter[i]==min){
					structuresWithLessDataItemsNotCorresponding.add(structures.get(i));
				}
			}
		}
		
		if (structuresWithLessDataItemsNotCorresponding.size()==1){ //somente 1 estrutura possui menos itens que nao correspondem a seus atributos
			resultingClasses.add(diagram.getClass(structuresWithLessDataItemsNotCorresponding.get(0).getName()));
		}
		else{ //duas ou mais estruturas possuem menos itens que nao correspondem a seus atributos
			
			//Procura os itens de dados que pertencem as estruturas,
			// entre as classes que possuem menos itens que nao correspondem a seus atributos
			int numberOfStructures2 = structuresWithLessDataItemsNotCorresponding.size();
			int[] dataItemsThatBelongCounter = new int[numberOfStructures2];
			Arrays.fill(dataItemsThatBelongCounter, 0);
			for (int i=0; i<numberOfStructures2; i++){
				List<DataItem> dataItems = structuresWithLessDataItemsNotCorresponding.get(i).getDataItems();
				for (DataItem dataItem : dataItems){
					if (diagram.getClass(structuresWithLessDataItemsNotCorresponding.get(i).getName()).containsAttribute(dataItem.getName())){
						dataItemsThatBelongCounter[i]++;
					}
				}
			}
			
			// Procurar as estruturas que possuem mais itens de dados correspondentes aos atributos da sua classe,
			// entre as classes que possuem menos itens que nao correspondem a seus atributos
			int max = 0;
			for (int i = 0; i<numberOfStructures2; i++){
				if (dataItemsThatBelongCounter[i]>max){
					max = dataItemsThatBelongCounter[i];
					resultingClasses.clear();
					resultingClasses.add(diagram.getClass(structuresWithLessDataItemsNotCorresponding.get(i).getName()));
				}
				else{
					if (dataItemsThatBelongCounter[i]==max){
						resultingClasses.add(diagram.getClass(structuresWithLessDataItemsNotCorresponding.get(i).getName()));
					}
				}
			}
			
			
		}
	}	


	private void mapDataItemsOfStructureToAssociations(ClassDiagram diagram, Structure structure){
		//Mtodo necessrio para mapear itens de dados presentes em estruturas que no correspondem 
		//s suas classes para associaes entre as classes dos itens de dados e as estruturas. (DIRETRIZ 3)
			try{
				br.ufsc.uid2uml.UMLDiagram.Class aClass = diagram.getClass(structure.getName());
				List<DataItem> dataItems = structure.getDataItems();
				for (int h = 0, sizeD = dataItems.size(); h < sizeD; h++){
					DataItem item = dataItems.get(h);
					if (!aClass.containsAttribute(item.getName())){
						//ENCONTRANDO A OUTRA CLASSE ASSOCIADA...
						List<br.ufsc.uid2uml.UMLDiagram.Class> classes = diagram.getClasses();
						for (int i = 0, size=classes.size(); i < size; i++){
							br.ufsc.uid2uml.UMLDiagram.Class anotherClass = classes.get(i);
 							if (anotherClass!=aClass && anotherClass.containsAttribute(item.getName())){   
								if (!aClass.containsAssociationWithClassAndSameAssociationName(anotherClass.getName(), aClass.getName()+"-"+anotherClass.getName())){
									Association a = diagram.insertAssociationBetweenClasses(aClass.getName()+"-"+anotherClass.getName(), 
											aClass.getName(), anotherClass.getName()); //aClass = source; anotherClass = target
									if (a != null){
										// Se o item de dado que representa o atributo a  um item de dado simples, 
										// ento a cardinalidade mxima da classe que contm a na associao  1. 
										// Se o item de dado que representa o atributo a  um conjunto, 
										// ento a cardinalidade mxima da classe que contm a na associao  N. 
										if (item.isSet()) //SET
											a.setTargetMaxCardinality(-1);
										else  					 //SINGLE
											a.setTargetMaxCardinality(1);
										// Se o item de dado que representa o atributo a  opcional, 
										// ento a cardinalidade mnima da classe que contm a na associao  0. 
										if (item.getCardinalityMin() == 0)
											a.setTargetMinCardinality(0);
									}
 							    }
							}
						}
					}
				}	
			}
			catch (ClassDiagramException e){
			}
			finally{
				List<Structure> internalStructures = structure.getStructures();
				for (int j = 0, sizeI = internalStructures.size(); j < sizeI; j++){
					Structure internalStruct = internalStructures.get(j);
					this.mapDataItemsOfStructureToAssociations(diagram, internalStruct);
				}
			}
		}

	private void mapInternalStructureToAssociations(ClassDiagram diagram, Structure structure){
		//Mtodo necessrio para mapear a estrutura interna a uma associao entre a classe correspondente
		// estrutura e a classe correspondente  estrutura interna. (DIRETRIZ 4)
			try{
				br.ufsc.uid2uml.UMLDiagram.Class aClass = diagram.getClass(structure.getName());				
				List<Structure> internalStructures = structure.getStructures();
				for (Structure internalStructure : internalStructures) {
					br.ufsc.uid2uml.UMLDiagram.Class anotherClass = diagram.getClass(internalStructure.getName());
					if (!aClass.containsAssociationWithClassAndSameAssociationName(anotherClass.getName(), aClass.getName()+"-"+anotherClass.getName())){
						Association a = diagram.insertAssociationBetweenClasses(aClass.getName()+"-"+anotherClass.getName(), 
								aClass.getName(), anotherClass.getName()); //aClass = source; anotherClass = target
						if (a != null){
						
							// Se a estrutura s2  uma estrutura simples, ou seja, no  um conjunto, 
							// ento a cardinalidade mxima da classe que corresponde a s2 na associao  1. 
							// Se a estrutura s2  um conjunto, 
							// ento a cardinalidade mxima da classe que corresponde a s2 na associao  N. 
							if (internalStructure.isSet()) //SET
								a.setTargetMaxCardinality(-1);
							else  					 //SINGLE
								a.setTargetMaxCardinality(1);
								
							// Se a estrutura s2  opcional, ento a cardinalidade mnima da classe que corresponde a s2 na associao  0. 
							if (internalStructure.isOptional())
								a.setTargetMinCardinality(0);						
						}
					}
				}
			}
			catch (ClassDiagramException e){
			}
		}

	
			
	/** This method save the class diagram in a file. 
	 *  The name of this file is the same of the class diagram with the
	 *  extension .diagram. 
	 * 
	 */
	public void saveIntoFile (){
	try{
		
		String fileName = this.classDiagramName + ".diagram";
		FileOutputStream fout = new FileOutputStream(fileName);
		PrintStream out = new PrintStream(fout);
		System.out.println("writing dictionary to file " + new File(fileName).getAbsolutePath());
		DictionaryEntry[] dictEntries =  this.classDiagram.getDictionaryEntries();
		for (int i=0; i<dictEntries.length; i++) {
			out.println("<dictionary-entry");
			out.println("    data item = " + dictEntries[i].dataItem);
			out.println("    owner = " + dictEntries[i].owner);
			out.println("    state = " + dictEntries[i].state);
			out.println("/>");
		}
		out.flush();
		out.close();
		System.exit(0);
	}
	catch (Exception e){
	}
	}	
	
	//FUNCOES AUXILIARES
	
	private List<String> structs (String nameItem) {
		//Funo que retorna uma lista com as estruturas onde um item aparece:
		//	structs (Item) = listStructures 
		//onde listStructures  uma lista contendo as estruturas nas quais o item de dado Item aparece.


		//list containing the name of all structures that contain the nameItem.
		List<String> structsName = new ArrayList<String>();

		// para cada UID, para cada State e SubState, adiciona o nome de todas estruturas, inclusive as internas
		//de outras estrutura, que ainda nao foram adicionadas.
		for (int i = 0; i < this.setUIDs.size(); i++) {
			UID uid = this.setUIDs.get(i);
			for (AbstractState s : uid.getSetStates()) {
				for (Structure str : s.getAllStructuresAndInternalStructures()) {
					if (str.containsDataItem(nameItem)) {
						if (!structsName.contains(str.getName())) {
							structsName.add(str.getName());
						}
					}
				}
			}
		}
		return structsName;
	}

	private int cardinality(String structure, String item) {
		
		// Funo que retorna a cardinalidade mxima de um atributo dentro de uma estrutura:
		// cardinality (Structure, Item) = X 
	    // onde X  a cardinalidade mxima do item de dado Item dentro da estrutura Structure. 
		// Se o item de dado Item no aparece na estrutura Structure, X  igual a 0. 
		// Se o item de dado Item aparece mais de uma vez na estrutura Structure, 
		// ento X deve retornar a menor destas cardinalidades mximas. 
		
		int x = 0;
		for (int i = 0; i < this.setUIDs.size(); i++) {
			UID uid = this.setUIDs.get(i);
			for (AbstractState s : uid.getSetStates()) {
				for (Structure str : s.getAllStructures()) {
					if (str.getName().equals(structure) && str.containsDataItem(item)) {
						DataItem d = str.getDataItem(item);
						if (x == 0 || x == -1) {
							x = d.getCardinalityMax();
						} 
						else 
							if (d.getCardinalityMax() < x) {
								x = d.getCardinalityMax();
							}
					}
				}
			}
		}
		return x;
	}

	private int times(String structure, String item) {
		
		//Funo que retorna o nmero de vezes que um item aparece junto com uma estrutura:
		//	times (Structure, Item) = X
		// onde X  o nmero de vezes em que o item de dado Item aparece dentro da estrutura Structure.

		int x = 0;
		for (int i = 0; i < this.setUIDs.size(); i++) {
			UID uid = this.setUIDs.get(i);
			for (AbstractState s : uid.getSetStates()) {
//				for (Structure str : s.getAllStructures()) {
				for (Structure str : s.getAllStructuresAndInternalStructures()) {
					if (str.getName().equals(structure)	&& str.containsDataItem(item)) {
						x++;
					}
				}
			}
		}
		return x;
	}

	
}

